#!/usr/bin/env python

import sys
from abipy.ppcodes.oncv_parser import OncvParser

parser = OncvParser(sys.argv[1])
parser.scan()
print(parser)

# To plot data with matplotlib.
plotter = parser.get_plotter()
#plotter.plot_atanlogder_econv()






# coding: utf-8
"""
Works and Flows for GWR calculations (GW with supercells).

NB: An Abinit build with Scalapack is required to run GWR.
"""
from __future__ import annotations

import os

from abipy.abio.inputs import AbinitInput, RUNL, GWR_TASK
from abipy.electrons.gwr import GwrRobot
from abipy.tools.iotools import make_executable
from .nodes import Node
from .tasks import TaskManager
from .works import Work


class DirectDiagoWork(Work):
    """
    This work performs the direct diagonalization of the KS Hamiltonian
    using the density produced by a GS-SCF run and produces a WFK file with
    empty states in the outdir of the second task.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DirectDiagoWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput, green_nband: int,
                       manager: TaskManager=None) -> DirectDiagoWork:
        """
        Build object from an input representing a GS-SCF calculation.

        Args:
            scf_input: Input for the GS-SCF calculation.
            green_nband: Number of bands to compute in the direct diagonalization.
                A negative value activate full diagonalization with nband equal to
                the number of PWs.
        """
        work = cls(manager=manager)
        work.green_nband = green_nband
        gwr_task = GWR_TASK.HDIAGO_FULL if green_nband < 0 else GWR_TASK.HDIAGO
        work.scf_task = work.register_scf_task(scf_input)
        diago_input = scf_input.new_with_vars(optdriver=RUNL.GWR, gwr_task=gwr_task)
        work.diago_task = work.register_gwr_task(diago_input, deps={work.scf_task: "DEN"})

        return work


class _BaseGWRWork(Work):

    @classmethod
    def from_varname_values(cls, varname_values: tuple, gwr_template: AbinitInput,
                            den_node: Node, wfk_node: Node,
                            manager: TaskManager = None):
        """
        Generate the work by changing the values of selected variables in a template for GWR calculations.

        Args:
            varname_values: Tuple or List of tuples. See examples below.
            gwr_template: GWR input to be used as template to generate other inputs.
            den_node: The Node who produced the DEN file.
            wfk_node: The Node who produced the WFK file.
            manager: Abipy Task Manager.

        Example:

            varname_values = ("nband", [8, 12, 14])
            Work.from_varname_values(varname_values, gwr_template, den_node, wfk_node)

            or:

            varname_values = [
                ("nband", [8, 12]),
                ("gwr_ntau", [6, 8]),
            ]

            Work.from_varname_values(varname_values, gwr_template, den_node, wfk_node)
        """
        work = cls(manager=manager)
        for new_inp in gwr_template.news_varname_values(varname_values):
            work.register_gwr_task(new_inp, deps={den_node: "DEN", wfk_node: "WFK"})

        return work


class GWRSigmaConvWork(_BaseGWRWork):
    """
    This work performs multiple QP calculations with the GWR code
    and produces `xlsx` files in its `outdata` directory
    with the QP results obtained with the different parameters.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GWRSigmaConvWork
    """

    def on_all_ok(self):
        """
        Generate `qp_dirgaps.csv` file in the `outdata` directory.
        """
        with GwrRobot.from_files(self.get_all_outdata_files_with_ext("_GWR.nc")) as gwr_robot:
            dirgaps_df = gwr_robot.get_dirgaps_dataframe()
            dirgaps_df.to_csv(self.outdir.path_in("dirgaps.csv"))
            qpdata_df = gwr_robot.get_dataframe()
            qpdata_df.to_csv(self.outdir.path_in("qpdata.csv"))

            with gwr_robot.get_pyscript(self.outdir.path_in("gwr_robot.py")) as script:
                script.add_text("""
#robot.plot_selfenergy_conv(spin=0, kpoint=?, band=?, axis="wreal", sortby=None, hue=None)
#robot.plot_qpgaps_convergence(qp_kpoints="all", qp_type="qpz0", sortby=None, hue=None)
""")

        return super().on_all_ok()



class GWRChiCompareWork(_BaseGWRWork):
    """
    This work computes the irreducibile polarizability along the imaginary axis
    using the GWR code and the quartic-scaling algorithm using the same minimax mesh
    so that one can compare the two quantities.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GWRChiCompareConvWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput, gwr_ntau, nband, ecuteps,
                       den_node: Node, wfk_node: Node,
                       gwr_kwargs=None, scr_kwargs=None,
                       manager: TaskManager = None):
        """
        Build Work from an input for GS-SCF calculation

        Args:
            scf_input: input for GS run.
            gwr_ntau: Number of points in minimax mesh.
            nband: Number of bands to build G and chi.
            ecuteps: Cutoff energy for chi.
            den_node: The Node who produced the DEN file.
            wfk_node: The Node who produced the WFK file.
            gwr_kwargs: Extra kwargs used to build the GWR input.
            scr_kwargs: Extra kwargs used to build the SCR input.
            manager: Abipy Task Manager.
        """
        gwr_input = scf_input.make_gwr_qprange_input(gwr_ntau=gwr_ntau, nband=nband,
                                                     ecuteps=ecuteps, gwr_task=GWR_TASK.CHI0)
        gwr_input.set_vars(prtsuscep=1, iomode=3)
        if gwr_kwargs is not None: gwr_input.set_vars(**gwr_kwargs)

        chi_input = scf_input.new_with_vars(optdriver=3,
                                            gwcalctyp=1, # Analytic continuation.
                                            nfreqim=gwr_ntau,
                                            ecuteps=ecuteps,
                                            nband=nband,
                                            prtsuscep=1,
                                            iomode=3,
                                            userie=4242, # Magic number to use the minimax mesh in the SCR driver.
                                                         # with nfreqim points.
                                            )
        if scr_kwargs is not None: chi_input.set_vars(**scr_kwargs)

        work = cls(manager=manager)
        work.register_scr_task(chi_input, deps={wfk_node: "WFK"})
        work.register_gwr_task(gwr_input, deps={den_node: "DEN", wfk_node: "WFK"})
        return work

    def on_all_ok(self):
        """
        Write python script to compare chi0 matrix elements.
        """
        sus_path = self[0].outdir.path_in("out_SUS.nc")
        tchi_path = self[1].outdir.path_in("out_TCHIM.nc")

        py_text = f"""\
#!/usr/bin/env python

import seaborn as sns
sns.set(context="paper", style='darkgrid', palette='deep',
       font='sans-serif', font_scale=0.8, color_codes=False, rc=None)

from abipy.electrons.gwr import TchimVsSus
o = TchimVsSus("{tchi_path}", "{sus_path}")

gpairs = [
    ((0, 0, 0), (0, 0, 0)),
    ((0, 0, 0), (1, 0, 0)),
    ((1, 0, 0), (1, 0, 0)),
    ((1, 0, 0), (0, 1, 0)),
    ((0, 0, 0), (2, 0, 0)),
]

qpoint_list = [
    [0, 0, 0],
    [0.5, 0, 0],
]

o.expose_qpoints_gpairs(qpoint_list, gpairs, exposer="mpl")

#o.plot_matdiff(qpoint=0, iw_index=0)
"""
        py_path = os.path.join(self.workdir, "plot_tchi_sus.py")
        with open(py_path, "wt") as fh:
            fh.write(py_text)
        make_executable(py_path)

        return super().on_all_ok()


class GWRRPAConvWork(_BaseGWRWork):
    """
    This work computes the RPA correlated energy for different number
    of points in the minimax mesh.
    """

    @classmethod
    def from_scf_input_ntaus(cls, scf_input: AbinitInput, gwr_ntau_list, nband, ecuteps,
                             den_node: Node, wfk_node: Node,
                             gwr_kwargs=None, manager: TaskManager=None):
        """
        Build Work from an input for GS-SCF calculation

        Args:
            scf_input: input for GS run.
            gwr_ntau_list: List with number of points in minimax mesh.
            nband: Number of bands to build G and chi.
            ecuteps: Cutoff energy for chi.
            den_node: The Node who produced the DEN file.
            wfk_node: The Node who produced the WFK file.
            gwr_kwargs: Extra kwargs used to build the GWR input.
            manager: Abipy Task Manager.
        """
        work = cls(manager=manager)
        for gwr_ntau in gwr_ntau_list:
            gwr_input = scf_input.make_gwr_qprange_input(gwr_ntau=gwr_ntau, nband=nband,
                                                         ecuteps=ecuteps, gwr_task=GWR_TASK.RPA_ENERGY)
            if gwr_kwargs is not None: gwr_input.set_vars(**gwr_kwargs)
            work.register_gwr_task(gwr_input, deps={den_node: "DEN", wfk_node: "WFK"})

        return work

    def on_all_ok(self):
        """
        Generate `gwr_robot.py` script in the `outdata` directory.
        """
        with GwrRobot.from_files(self.get_all_outdata_files_with_ext("_GWR.nc")) as gwr_robot:
            with gwr_robot.get_pyscript(self.outdir.path_in("gwr_robot.py")) as script:
                script.add_text("""
df = robot.get_rpa_ene_dataframe()
print(df)
""")

        return super().on_all_ok()



#!/usr/bin/env python
r"""
GWR with/without supercell for Sigma
====================================

This script computes the G0W0 corrections with the GWR code
using real-space supercell or the mixed space algorithm.
"""

import os
import sys
import abipy.data as data
import abipy.abilab as abilab
import abipy.core.abinit_units as abu

from abipy import flowtk


def build_flow(options):

    from abipy.data.gwr_structures import get_gwr_structure
    symbol = options.extra
    if symbol is None:
        raise ValueError("--extra argument with structure symbol is needed")
    structure = get_gwr_structure(symbol)

    options.workdir = symbol

    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    #if not options.workdir:
    #    options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_","flow_")

    from abipy.flowtk.psrepos import get_repo_from_name
    pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("standard")
    #pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("stringent")

    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)
    scf_input.set_cutoffs_for_accuracy("normal")
    ecut = scf_input["ecut"]
    scf_input.set_scf_nband_semicond()

    # Global variables.
    scf_input.set_vars(
        tolvrs=1e-8,
        paral_kgb=0,
        npfft=1,
        timopt=-1,
    )
    scf_input.set_kmesh(
        #ngkpt=[1, 1, 1],
        #ngkpt=[2, 2, 2],
        #ngkpt=[4, 4, 4],
        ngkpt=[6, 6, 6],
        shiftk=[0.0, 0.0, 0.0], # IMPORTANT: k-grid for GWR must be Gamma-centered.
    )

    # Get max number of PWs.
    #dims, _ = scf_input.abiget_dims_spginfo()
    #mpw = dims["mpw"]

    flow = flowtk.Flow(workdir=options.workdir)
    small_manager = options.manager.new_with_fixed_mpi_omp(4, 1)

    # GS-SCF run to get the DEN, followed by direct diago to obtain green_nband bands.
    from abipy.flowtk.gwr_works import DirectDiagoWork, GWRSigmaConvWork
    green_nband = -1  # -1 this means full diago
    diago_work = DirectDiagoWork.from_scf_input(scf_input, green_nband)
    diago_work[0].set_manager(small_manager)
    flow.register_work(diago_work)

    # Build template for GWR.
    ecuteps = 12
    nval = scf_input.num_valence_electrons // 2
    nband = nval * 100
    gwr_template = scf_input.make_gwr_qprange_input(gwr_ntau=6, nband=nband, ecuteps=ecuteps)

    # Define kptgw and bdgw
    kptgw = [ # k-points in reduced coordinates
        (0.0, 0.0, 0.0),
        (0.5, 0.5, 0.0), # X
        #(0.5    0.000    0.000),
    ]

    nkptgw = len(kptgw)
    bdgw = (nval, nval+1) * nkptgw

    sigma_kcalc_dict = dict(
        nkptgw=nkptgw,
        kptgw=kptgw,
        bdgw=bdgw,
    )

    gwr_template.set_vars(**sigma_kcalc_dict)

    gwr_ntau_list = [6]
    #gwr_ntau_list = [20]

    # Conpute QP corrections without/with regularization term.
    # 1) Change the value of one variable:
    varname_values = ("gwr_ntau", gwr_ntau_list,
                      "gwr_sigma_algo_list": [1, 2],
                     )
    gwr_template["nband"] = nband
    gwr_template["gwr_regterm"] = 1e-6
    #gwr_template["gwr_regterm"] = 0.0

    wfk_node = diago_work[1]
    gwr_work = GWRSigmaConvWork.from_varname_values(
            varname_values, gwr_template, den_node=diago_work[0], wfk_node=wfk_node)
    flow.register_work(gwr_work)

    return flow


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
This script provides a command line interface to a MongoDB collection containing FlowModels
"""
from __future__ import annotations

import sys
import os
import ast
import argparse

from typing import Type
from pprint import pprint  #, pformat
from pydantic import BaseModel
from pymongo.collection import Collection
from monty.functools import prof_main
from monty.termcolor import cprint
from abipy.core.release import __version__
#from abipy.core.config import get_config
from abipy.core.structure import Structure
from abipy.tools.printing import print_dataframe
from abipy.htc.base_models import MongoConnector
from abipy.htc.flow_models import FlowModel
from abipy.htc.protocol import Protocol


MAGIC_COLLECTION_FILENAME = "_abidb_collection"


def get_epilog() -> str:
    usage = f"""\

Usage example:

  abidb.py list
  abidb.py stats -c COLLECTION_NAME

  abidb.py init -c COLLECTION_NAME -f FLOW_MODEL_CLASS -p PROTOCOL_NAME
  abidb.py add  -c COLLECTION_NAME -p PROTOCOL_NAME si.cif gaas.cif


TIP: If you are using a single collection and you don't want to specify the collection name with `-c`
     you can create a `{MAGIC_COLLECTION_FILENAME}` file in the directory in which the abidb.py script is executed.
     This file should contain a single line with the name of the MongoDB collection.
"""

    return usage


def _parse_dict(string):
    obj = ast.literal_eval(string)  #; print(obj)
    if not isinstance(obj, dict):
        raise TypeError(f"Expecting dictionary, got {type(obj)}\nobj: {obj}")
    return obj


def _parse_dict_or_list(string):
    obj = ast.literal_eval(string)  #; print(obj)
    if not isinstance(obj, (dict, list)):
        raise TypeError(f"Expecting dictionary or list, got {type(obj)}\nobj: {obj}")
    return obj


def get_parser(with_epilog=False):

    # Parent parser for commands requiring a `worker_name`
    #worker_selector = argparse.ArgumentParser(add_help=False)
    #worker_selector.add_argument('worker_name', type=str, help="Name of the worker server.")

    # Parent parser for common options.
    copts_parser = argparse.ArgumentParser(add_help=False)
    copts_parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
        help='verbose, can be supplied multiple times to increase verbosity.')

    copts_parser.add_argument('--loglevel', default="ERROR", type=str,
                              help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG.")

    copts_parser.add_argument("-c", "--collection-name", type=str,
                               default="",
                               #required=True,
                               help="Name of the MongDB collection used to store FlowModels.")

    copts_parser.add_argument("--local-db", action='store_true', default=False,
                              help="Assume MongoDB server is running on the localhost with the default port.")

    #serve_parser = argparse.ArgumentParser(add_help=False)
    #serve_parser.add_argument("-a", "--address", type=str, default="localhost", help="Address")
    #serve_parser.add_argument("-p", "--port", type=int, default=0,
    #                          help="Port. Port 0 means to select an arbitrary unused port")
    #serve_parser.add_argument("-d", "--daemonize", action="store_true", default=False, help="Demonize the server")

    # Build the main parser.
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    # Create the parsers for the sub-commands
    subparsers = parser.add_subparsers(dest='command', help='sub-command help', description="Valid subcommands")

    p_list = subparsers.add_parser("list", parents=[copts_parser], help=abidb_list.__doc__)

    p_init = subparsers.add_parser("init", parents=[copts_parser], help=abidb_init.__doc__)
    p_init.add_argument("-p", "--protocol", type=str, default=None,
                        help="Protocol name or path to Yaml file with protocol. "
                             "Empty to print all avaiable protocols."
                        )
    p_init.add_argument("-f", "--flow-model-name", type=str, default=None,
                        help="Name of FlowModel class. Empty to print all available models."
                        )

    p_add = subparsers.add_parser("add", parents=[copts_parser], help=abidb_add.__doc__)
    p_add.add_argument("files", nargs="+", help="list of files from which structures are read.")
    p_add.add_argument("-mp", action="store_true", default=False,
                        help="Interpret files as list materials project IDs instead of filepaths")

    p_status = subparsers.add_parser("status", parents=[copts_parser], help=abidb_status.__doc__)

    p_aggreg = subparsers.add_parser("aggreg", parents=[copts_parser], help=abidb_aggreg.__doc__)

    p_flow_data = subparsers.add_parser("flow_data", parents=[copts_parser], help=abidb_flow_data.__doc__)
    p_flow_data.add_argument("oid", type=str, help="Document ID")

    p_preq = subparsers.add_parser("preq", parents=[copts_parser], help=abidb_preq.__doc__)
    p_preq.add_argument("pre_inds", nargs="*",
                      help="Index or indices of query."
                           "If not given, the full list of available queries is returned."
                           "If `all` all registered queries are executed."
                            )

    p_query = subparsers.add_parser("query", parents=[copts_parser], help=abidb_query.__doc__)
    p_query.add_argument("-q", "--query", type=_parse_dict, help="Pymongo query to filter entries")
    p_query.add_argument('-p', "--projection", type=_parse_dict_or_list, default=None, help="Pymongo projection")
    p_query.add_argument("-l", "--limit", type=int, default=0,
                         help="List number of documents returned by the query, 0 means ALL docs matching the query!")

    p_schema = subparsers.add_parser("schema", parents=[copts_parser], help=abidb_schema.__doc__)

    p_drop = subparsers.add_parser("drop", parents=[copts_parser], help=abidb_drop.__doc__)

    p_show_err = subparsers.add_parser("show_err", parents=[copts_parser], help=abidb_show_err.__doc__)

    #p_info_oid = subparsers.add_parser("info_oid", parents=[copts_parser], help=abidb_info_oid.__doc__)

    #p_mongo_start = subparsers.add_parser(
    #    "mongo_start", parents=[copts_parser, serve_parser, worker_selector_with_default],
    #    help="Create and start new AbiPy worker connected to a MongoDB database containing FlowModels.")
    #p_mongo_start.add_argument("collection_name", type=str,
    #p_mongo_start.add_argument("-c", "--collection-name", type=str,
    #                           default="",
    #                           #required=True,
    #                           help="Name of the MongDB collection used to store FlowModels.")
    #p_mongo_start.add_argument("-s", "--scratch-dir", type=str, default=None,
    #                           help="Scratch directory in which Flows will be generated. "
    #                                "If not given, use value from ~/.abinit/abipy/config.yml")

    p_mongo_gui = subparsers.add_parser("mongo_gui", parents=[copts_parser], # , serve_parser],
                                        help="Start MongoGUI.")

    return parser


#def tools():
#    from pymongo import DESCENDING, ASCENDING
#    if hasattr(args, "sort") and args.sort:
#        sort = [(args.sort, ASCENDING)]
#    elif hasattr(args, "rsort") and args.rsort:
#        sort = [(args.rsort, DESCENDING)]
#    else:
#        sort = None


def serve_kwargs_from_options(options):

    print("""
Use:

    ssh -N -f -L localhost:{port}:localhost:{port} username@your_remote_cluster

to activate port forwarding.
""")

    import abipy.panels as mod
    assets_path = os.path.join(os.path.dirname(mod.__file__), "assets")

    d = dict(
        debug=options.verbose > 0,
        #show=False,
        #port=options.port,
        static_dirs={"/assets": assets_path},
        #address=options.address,
        websocket_origin="*",
        panel_template="FastList",
    )
    #print("serve_kwargs:\n", pformat(d), "\n")
    return d


#def serve(worker, options):
#    s = str(worker) if options.verbose else repr(worker)
#    print(s)
#
#    serve_kwargs = serve_kwargs_from_options(options)
#
#    if not options.daemonize:
#        worker.serve(**serve_kwargs)
#    else:
#        print(f"Running server {worker.name} in demon mode...")
#        import daemon
#        with daemon.DaemonContext():
#            worker.serve(**serve_kwargs)
#
#    return 0


class Context(BaseModel):

    mng_connector: MongoConnector

    collection: Collection

    flow_model_cls: Type[FlowModel]

    class Config:
        arbitrary_types_allowed = True


def abidb_list(options, ctx: Context):
    """
    List MongoDB collections.
    """
    coll_names = ctx.mng_connector.list_collection_names()
    coll_names = [cn for cn in coll_names if not cn.startswith("gridfs_")]
    print("List of MongoDB collections:\n")
    for cname in coll_names:
        print("- ", cname)

    return 0


#def abidb_info(options, ctx: Context):


def abidb_init(options, mng_connector):  #, ctx: Context):
    """
    Initialize a MongoDB collection to perform calculations with a given FlowModel and (default) protocol.
    """
    if not options.protocol:
        print("Printing all registered protocols and then exit\n")
        for p in Protocol.get_all_abipy_protocols():
            #print(p)
            print(repr(p))
        return 0

    elif os.path.exists(options.protocol):
        # Read protocol from an user-provided Yaml file
        protocol = Protocol.from_file(options.protocol)

    else:
        # Get official protocol from its name
        protocol = Protocol.from_name(options.protocol)

    print("Using protocol:\n", protocol)
    #print(protocol.info)
    #from abipy.htc import get_all_flow_models_with_protocol

    if not options.flow_model_name:
        print("Printing all FlowModelWithProtocol and exit as flow_model_name is not given...")
        #for flow_model_cls in get_all_flow_model_cls_with_protocol()
        #   print(flow_mode_cls)
        return 0

    #flow_model_cls = get_all_flow_model_cls_with_protocol(cls_name=options.flow_model_name)[0]
    #print(f"Using {flow_model_cls}")
    #mng_connector.init_flow_model_collection(flow_model_cls, protocol=protocol)

    return 0


def abidb_add(options, mng_connector):  #, ctx: Context):
    """
    Convenient command line interface to add FlowModel object to a MongoDB collection
    starting from a list of external files with structural info or mp-ids.
    """
    if options.mp:
        # Get structures from the Materials Project website via MP ids.
        structures = [Structure.from_mpid(mpid) for p in options.files]
    else:
        # Get structures from external files.
        structures = [Structure.from_file(p) for p in options.files]

    # Retrieve class and protocol from collection.
    #flow_model_cls, protocol = mng_connector.get_flow_model_cls_and_protocol()

    # Allow the user to change the protocol
    if options.protocol:
        if os.path.exists(options.protocol):
            # Read protocol from user-provided Yaml file.
            new_protocol = Protocol.from_file(options.protocol)
        else:
            # Get official protocol from name.
            new_protocol = Protocol.from_name(options.protocol)

        print("replacing initial protocol with new_protocol")
        protocol = new_protocol

    print("Using protocol:")
    print(protocol)

    flow_models = []
    for structure in structures:
        m = flow_model_cls.from_structure_and_protocol(structure, protocol)
        flow_models.append(m)

    # Inser FlowModels in the collection
    mng_connector.insert_flow_models(flow_models, verbose=1)

    return 0


def abidb_query(options, ctx: Context):
    print("query", options.query)
    print("projection", options.projection)
    print("limit", options.limit)

    for doc in ctx.collection.find(options.query, projection=options.projection, limit=options.limit):
        print(doc)

    return 0


def abidb_status(options, ctx: Context):
    status2oids = ctx.flow_model_cls.mng_get_status2oids(ctx.collection)
    for status, oids in status2oids.items():
        print(str(status), len(oids))

    return 0


def abidb_aggreg(options, ctx: Context):
    # Find all "mng_aggregate_" class method
    import inspect
    def predicate(value):
        return inspect.ismethod(value) and value.__name__.startswith("mng_aggregate")

    for name, value in inspect.getmembers(ctx.flow_model_cls, predicate):
        print(name, value)
        df = value(ctx.collection)
        print_dataframe(df)

    #df = ctx.flow_model_cls.mng_aggregate_in_struct(ctx.collection)

    return 0


def abidb_flow_data(options, ctx: Context):
    flow_data = ctx.flow_model_cls.mongo_get_flow_data(options.oid, ctx.collection)
    #print(repr(flow_data))
    #print(flow_data.yaml_dump())
    flow_data.summarize(verbose=options.verbose)
    #for work_data in flow_data.works_data:
    #    print(work_data)

    return 0


def abidb_preq(options, ctx: Context):
    """
    Execute preset queries on the given collection.
    """
    queries = ctx.flow_model_cls.get_preset_queries()

    if not options.preq_inds:
        print("Printing list of predefined queries with their index as `cg_inds` argument is not given\n")
        for i, query in enumerate(queries):
            print(query.to_string(title=f"== Index: {i} ===", verbose=options.verbose))

        print("Use e.g. `abidb.py preq 0 1` to execute queries with a given index")
        print("      or `abidb.py preq all` to select all")
        print("Use -v to print extra info")
        return 0

    cg_inds = set(options.preq_inds)
    if "all" in cg_inds:
        cg_inds = set(range(len(queries)))
    else:
        cg_inds = set([int(c) for c in cg_inds])

    for i, query in enumerate(queries):
        if i not in cg_inds: continue
        print(query.to_string(verbose=options.verbose))
        df = query.get_dataframe(ctx.collection, try_short_keys=True)
        print(df)
        #print_dataframe(df)

    return 0


def abidb_schema(options, ctx: Context):
    """
    Show FlowModel schema.
    """
    #print(flow_model_cls)
    #print(flow_model_cls.schema_json(indent=2))
    schema = ctx.flow_model_cls.schema()
    import ruamel.yaml as yaml
    print(yaml.safe_dump(schema, default_flow_style=False))
    return 0


def abidb_drop(options, ctx: Context):
    """
    Drop the specified MongoDB collection and the associated GridFs collections (use it wisely!)
    """
    return ctx.mng_connector.drop_collection(ask_for_confirmation=True)


def abidb_show_err(options, ctx: Context):
    """
    Show errored Flows
    """
    cnt = 0
    #for flow_model in ctx.flow_model_cls.find_err_flow_models(ctx.mng_connector):
    for (oid, flow_model) in ctx.mng_connector.find_err_oid_flowmodels():
        cnt += 1
        print(f"Traceback for FlowModel: `{flow_model.__class__.__name__}` with oid: {oid}")
        for t in flow_model.flow_data.tracebacks:
            cprint(t, color="red")

    return 0


def abidb_mongo_gui(options, ctx: Context):
    """
    Start GUI.
    """
    serve_kwargs = serve_kwargs_from_options(options)
    ctx.mng_connector.open_mongoflow_gui(flow_model_cls=ctx.flow_model_cls, **serve_kwargs)


@prof_main
def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg: sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse command line.
    try:
        options = parser.parse_args()
    except Exception as exc:
        raise exc
        print(exc)
        show_examples_and_exit(error_code=1)

    if not options.command:
        print("command argument is required!")
        show_examples_and_exit(error_code=1)

    # loglevel is bound to the string value obtained from the command line argument.
    # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
    import logging
    numeric_level = getattr(logging, options.loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % options.loglevel)
    logging.basicConfig(level=numeric_level)

   #try:
   #     import argcomplete
   #     argcomplete.autocomplete(parser)
   #     # This supports bash autocompletion. To enable this:
   #     #
   #     #   pip install argcomplete,
   #     #   activate-global-python-argcomplete
   #     #
   #     #  or add
   #     #      eval "$(register-python-argcomplete abidb)"
   #     #
   #     # into your .bash_profile or .bashrc
   # except ImportError:
   #     pass

    if options.local_db:
        mng_connector = MongoConnector.for_localhost(collection_name=options.collection_name)
    else:
        mng_connector = MongoConnector.from_abipy_config(collection_name=options.collection_name)
    print(mng_connector)

    if options.command in ("init", "add"):
        return globals()[f"abidb_{options.command}"](options, mng_connector) #, ctx)

    if not options.collection_name:
        # Take it from external file
        print(f"Taking collection_name from: {MAGIC_COLLECTION_FILENAME} as `-c` option is not used")
        try:
            with open(MAGIC_COLLECTION_FILENAME, "rt") as fh:
                options.collection_name = fh.read().strip()
                print(f"collection_name set to: {options.collection_name}")

        except FileNotFoundError:
            raise RuntimeError(
                  f"\nCannot find `{MAGIC_COLLECTION_FILENAME}`\n"
                  f"If you don't want to specify the collection name on the command line with `-c COLL_NAME`\n",
                  f"you need to create a `{MAGIC_COLLECTION_FILENAME}` file in the current working directory\n"
                  f"with one line giving the name of the MongoDB collection that should be used."
                  )

    collection = mng_connector.get_collection()
    flow_model_cls = FlowModel.get_subclass_from_collection(collection)

    ctx = Context(**locals())

    #config = get_config()
    #if options.verbose > 2:
    #    print(options)
    #    print("global abipy config options:\n", config)

    return globals()[f"abidb_{options.command}"](options, ctx)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""
Interface to the ESKW.nc file storing the (star-function) interpolated band structure produced by Abinit.
"""
from monty.functools import lazy_property
from monty.string import marquee
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.electrons.ebands import ElectronsReader


class EskwFile(AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    This file contains the (star-function) interpolated band structure produced by Abinit.
    It's similar to the GSR file but it does not contain the header and energies.

    Usage example:

    .. code-block:: python

        with EskwFile("foo_ESKW.nc") as eskw:
            eskw.ebands.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: EskwFile
    """
    @classmethod
    def from_file(cls, filepath: str):
        """Initialize the object from a netcdf_ file."""
        return cls(filepath)

    def __init__(self, filepath):
        super().__init__(filepath)
        self.reader = ElectronsReader(filepath)

    @lazy_property
    def einterp(self):
        return self.reader.read_value("einterp")

    @lazy_property
    def band_block(self):
        # band_block(2)=Initial and final band index to be interpolated. [0, 0] if all bands are used.
        band_block = self.reader.read_value("band_block")
        if all(band_block != [0, 0]): band_block -= 1
        return band_block

    def __str__(self):
        """String representation."""
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))
        app("band_block: %s" % str(self.band_block))
        app("einterp: %s" % str(self.einterp))

        return "\n".join(lines)

    def close(self):
        self.reader.close()

    @property
    def ebands(self):
        """|ElectronBands| object."""
        return self.reader.read_ebands()

    @property
    def structure(self):
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def params(self):
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        od["einterp"] = self.interp
        od["einterp"] = self.einterp
        return od

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        #for fig in self.yield_structure_figs(**kwargs): yield fig
        for fig in self.yield_ebands_figs(**kwargs): yield fig

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("eskw = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(eskw)"),
            nbv.new_code_cell("eskw.ebands.plot();"),
            nbv.new_code_cell("eskw.ebands.kpoints.plot();"),
            nbv.new_code_cell("# eskw.ebands.plot_transitions(omega_ev=3.0, qpt=(0, 0, 0), atol_ev=0.1);"),
            nbv.new_code_cell("""\
if eskw.ebands.kpoints.is_ibz:
    eskw.ebands.get_edos().plot();"""),
        ])

        return self._write_nb_nbpath(nb, nbpath)



# flake8: noqa
import os
import tempfile

from monty.termcolor import cprint
from pymatgen.io.abinit.abiobjects import *
from pymatgen.io.abinit.pseudos import Pseudo, PseudoTable, PseudoParser
from pymatgen.io.abinit.netcdf import NetcdfReader

from .abiinspect import GroundStateScfCycle, D2DEScfCycle, yaml_read_kpoints, yaml_read_irred_perts
from .launcher import PyFlowScheduler, PyLauncher
from .qadapters import show_qparams, all_qtypes
from .wrappers import Mrgscr, Mrgddb, Cut3D, Fold2Bloch
from .nodes import Status
from .tasks import *
from .tasks import EphTask, ElasticTask
from .works import *
from .flows import (Flow, G0W0WithQptdmFlow, bandstructure_flow, PhononFlow, phonon_conv_flow,
    g0w0_flow, NonLinearCoeffFlow)
from .abitimer import AbinitTimerParser, AbinitTimerSection
from .events import EventsParser, autodoc_event_handlers
#from abipy.flowtk.works import *
#from abipy.flowtk.gs_works import EosWork
from .dfpt_works import ElasticWork, NscfDdksWork
from .gw_works import ScreeningWork


def flow_main(main):  # pragma: no cover
    """
    This decorator is used to decorate main functions producing `Flows`.
    It adds the initialization of the logger and an argument parser that allows one to select
    the loglevel, the workdir of the flow as well as the YAML file with the parameters of the `TaskManager`.
    The main function shall have the signature:

        main(options)

    where options in the container with the command line options generated by `ArgumentParser`.

    Args:
        main: main function.
    """
    from functools import wraps

    @wraps(main)
    def wrapper(*args, **kwargs):
        # Build the parse and parse input args.
        parser = build_flow_main_parser()
        options = parser.parse_args()

        # loglevel is bound to the string value obtained from the command line argument.
        # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
        import logging
        numeric_level = getattr(logging, options.loglevel.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % options.loglevel)
        logging.basicConfig(level=numeric_level)

        # Istantiate the manager.
        options.manager = TaskManager.as_manager(options.manager)

        if options.tempdir:
            options.workdir = tempfile.mkdtemp()
            print("Working in temporary directory", options.workdir)

        def execute():
            """This is the function that performs the work depending on options."""
            flow = main(options)

            if options.plot:
                flow.plot_networkx(tight_layout=True, with_edge_labels=True)

            if options.graphviz:
                graph = flow.get_graphviz() #engine=options.engine)
                directory = tempfile.mkdtemp()
                print("Producing source files in:", directory)
                graph.view(directory=directory, cleanup=False)

            if options.abivalidate:
                print("Validating flow input files...")
                isok, errors = flow.abivalidate_inputs()
                if not isok:
                    for e in errors:
                        if e.retcode == 0: continue
                        lines = e.log_file.readlines()
                        i = len(lines) - 50 if len(lines) >= 50 else 0
                        print("Last 50 line from log file:")
                        print("".join(lines[i:]))
                    raise RuntimeError("flow.abivalidate_input failed. See messages above.")
                else:
                    print("Validation succeeded")

            if options.remove and os.path.isdir(flow.workdir):
                print("Removing old directory:", flow.workdir)
                import shutil
                shutil.rmtree(flow.workdir)

            if options.dry_run:
                print("Dry-run mode.")
                retcode = 0
            elif options.scheduler:
                retcode = flow.make_scheduler().start()
                if retcode == 0:
                    retcode = 0 if flow.all_ok else 1
            else:
                # Default behaviour.
                retcode = flow.build_and_pickle_dump()

            cprint("Return code: %d" % retcode, "red" if retcode != 0 else "green")
            return retcode

        if options.prof:
            # Profile execute
            import pstats, cProfile
            cProfile.runctx("execute()", globals(), locals(), "Profile.prof")
            s = pstats.Stats("Profile.prof")
            s.strip_dirs().sort_stats("time").print_stats()
            return 0
        else:
            return execute()

    return wrapper


def build_flow_main_parser():
    """
    Build and return the parser used in the abipy/data/runs scripts.
    """

    epilog = """\
======================================================================================================
Usage example:

    script.py                           => Build flow without submitting it.
    script.py -s                        => Build flow and start the scheduler
    nohup script.py -s > log 2> err &   => Build flow, start the scheduler in background
                                           This is the recommended approach for long-running calculations.
                                           Thanks to `nohup`, it is possible to disconnect the terminal
                                           without killing the scheduler.

======================================================================================================
"""

    import argparse
    parser = argparse.ArgumentParser(epilog=epilog,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--loglevel', default="ERROR", type=str,
                        help="set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")
    parser.add_argument("-w", '--workdir', default="", type=str, help="Working directory of the flow.")
    parser.add_argument("-m", '--manager', default=None,
                        help="YAML file with the parameters of the task manager. "
                             "Default None i.e. the manager is read from standard locations: "
                             "working directory first then ~/.abinit/abipy/manager.yml.")
    parser.add_argument("-s", '--scheduler', action="store_true", default=False,
                        help="Run the flow with the scheduler")
    parser.add_argument("-r", "--remove", default=False, action="store_true", help="Remove old flow workdir (if any)")
    parser.add_argument("-p", "--plot", default=False, action="store_true", help="Plot flow with networkx.")
    parser.add_argument("-g", "--graphviz", default=False, action="store_true", help="Plot flow with graphviz.")
    parser.add_argument("-d", "--dry-run", default=False, action="store_true", help="Don't write directory with flow.")
    parser.add_argument("-a", "--abivalidate", default=False, action="store_true", help="Call Abinit to validate input files.")
    parser.add_argument("-t", "--tempdir", default=False, action="store_true", help="Execute flow in temporary directory.")
    parser.add_argument("--prof", action="store_true", default=False, help="Profile code wth cProfile ")
    parser.add_argument("-e", "--extra", default=None, help="Extra argument passed to the script.")

    return parser



"""Integration tests for structural relaxations."""
from __future__ import annotations

import numpy as np
import pytest
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk

from abipy.core.testing import has_matplotlib


def ion_relaxation(tvars, ntime=50):
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))

    # Perturb the structure (random perturbation of 0.1 Angstrom)
    structure.perturb(distance=0.02)

    global_vars = dict(
        ecut=6,
        ngkpt=[2,2,2],
        shiftk=[0,0,0],
        nshiftk=1,
        chksymbreak=0,
        paral_kgb=tvars.paral_kgb,
    )

    inp = abilab.AbinitInput(structure, pseudos=abidata.pseudos("14si.pspnc"))

    # Global variables
    inp.set_vars(global_vars)

    # Dataset 1 (Atom Relaxation)
    #inp[1].set_vars(
    # FIXME here there's a bug
    inp.set_vars(
        optcell=0,
        ionmov=2,
        tolrff=0.02,
        tolmxf=5.0e-5,
        ntime=ntime,
        #dilatmx=1.05, # FIXME: abinit crashes if I don't use this
    )

    return inp


def itest_atomic_relaxation(fwp, tvars):
    """Test atomic relaxation with automatic restart."""
    # Build the flow
    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)

    ion_input = ion_relaxation(tvars, ntime=2)
    ion_input["chksymtnons"] = 0
    work = flow.register_task(ion_input, task_class=flowtk.RelaxTask)

    flow.allocate()
    flow.build_and_pickle_dump(abivalidate=True)

    # Run t0, and check status
    t0 = work[0]
    t0.start_and_wait()
    assert t0.returncode == 0
    t0.check_status()
    assert t0.status == t0.S_UNCONVERGED

    assert t0.initial_structure == ion_input.structure
    unconverged_structure = t0.get_final_structure()
    assert unconverged_structure != t0.initial_structure

    # Use the default value ntime=50 and we can converge the calculation.
    t0.input.set_vars(ntime=50)

    t0.build()
    assert t0.restart()
    t0.wait()
    assert t0.num_restarts == 1

    # At this point, we should have reached S_OK.
    assert t0.status == t0.S_DONE
    t0.check_status()
    assert t0.status == t0.S_OK

    final_structure = t0.get_final_structure()
    assert final_structure != unconverged_structure

    flow.show_status()
    assert all(work.finalized for work in flow)
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    # post-processing tools
    if has_matplotlib():
        assert t0.inspect(show=False) is not None

    with t0.open_hist() as hist:
        hist.to_string(verbose=2)
        # from_file accepts HIST files as well.
        assert np.all(hist.structures[-1].frac_coords == abilab.Structure.from_file(hist.filepath).frac_coords)

    with t0.open_gsr() as gsr:
        gsr.to_string(verbose=2)
        gsr.pressure == 1.8280

    t0.get_results()


def make_ion_ioncell_inputs(tvars, dilatmx, scalevol=1, ntime=50):
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))

    # Perturb the structure (random perturbation of 0.1 Angstrom)
    #structure.perturb(distance=0.01)

    # Compress the lattice so that ABINIT complains about dilatmx
    structure.scale_lattice(structure.volume * scalevol)

    global_vars = dict(
        ecut=6,
        ecutsm=0.5,
        ngkpt=[4,4,4],
        shiftk=[0,0,0],
        nshiftk=1,
        chksymbreak=0,
        paral_kgb=tvars.paral_kgb,
    )

    multi = abilab.MultiDataset(structure, pseudos=abidata.pseudos("14si.pspnc"), ndtset=2)

    # Global variables
    multi.set_vars(global_vars)

    # Dataset 1 (Atom Relaxation)
    multi[0].set_vars(
        optcell=0,
        ionmov=2,
        tolrff=0.02,
        tolmxf=5.0e-5,
        ntime=ntime,
    )

    # Dataset 2 (Atom + Cell Relaxation)
    multi[1].set_vars(
        optcell=1,
        ionmov=2,
        dilatmx=dilatmx,
        tolrff=0.02,
        tolmxf=5.0e-5,
        strfact=100,
        ntime=ntime,
    )

    ion_inp, ioncell_inp = multi.split_datasets()
    return ion_inp, ioncell_inp


def itest_relaxation_with_restart_from_den(fwp, tvars):
    """Test structural relaxations with automatic restart from DEN files."""
    # Build the flow
    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)

    # Use small value for ntime to trigger restart, then disable the output of the WFK file.
    ion_input, ioncell_input = make_ion_ioncell_inputs(tvars, dilatmx=1.1, ntime=3)
    ion_input.set_vars(prtwf=0)
    ioncell_input.set_vars(prtwf=0)

    relax_work = flowtk.RelaxWork(ion_input, ioncell_input)
    flow.register_work(relax_work)

    assert flow.make_scheduler().start() == 0
    flow.show_status()

    assert all(work.finalized for work in flow)
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    # we should have (0, 1) restarts and no WFK file in outdir.
    for i, task in enumerate(relax_work):
        assert task.status == task.S_OK
        assert task.num_restarts == i
        assert task.num_corrections == 0
        assert not task.outdir.has_abiext("WFK")

    if has_matplotlib:
        assert relax_work.plot_ion_relaxation(show=False) is not None
        assert relax_work.plot_ioncell_relaxation(show=False) is not None

    flow.rmtree()


def itest_dilatmx_error_handler(fwp, tvars):
    """
    Test cell relaxation with automatic restart in the presence of dilatmx error.
    """
    pytest.xfail("dilatmxerror_handler is not portable and it's been disabled!")
    # Build the flow
    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)

    # Decrease the volume to trigger DilatmxError
    ion_input, ioncell_input = make_ion_ioncell_inputs(tvars, dilatmx=1.01, scalevol=0.8)

    work = flowtk.Work()
    work.register_relax_task(ioncell_input)

    flow.register_work(work)
    flow.allocate()
    assert flow.make_scheduler().start() == 0
    flow.show_status()

    assert all(work.finalized for work in flow)
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    # t0 should have reached S_OK, and we should have DilatmxError in the corrections.
    t0 = work[0]
    assert t0.status == t0.S_OK
    print(t0.corrections)
    assert t0.num_corrections > 0
    assert t0.corrections[0]["event"]["@class"] == "DilatmxError"


def itest_relaxation_with_target_dilatmx(fwp, tvars):
    """Test structural relaxations with automatic restart from DEN files."""
    # Build the flow
    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)

    # Use small value for ntime to trigger restart, then disable the output of the WFK file.
    ion_input, ioncell_input = make_ion_ioncell_inputs(tvars, dilatmx=1.05)

    target_dilatmx = 1.03
    relax_work = flowtk.RelaxWork(ion_input, ioncell_input, target_dilatmx=target_dilatmx)
    flow.register_work(relax_work)

    assert flow.make_scheduler().start() == 0
    flow.show_status()

    assert all(work.finalized for work in flow)
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()
    #assert relax_work.last_dilatmx <= target_dilatmx

    # we should have (0, 1) restarts
    for i, task in enumerate(relax_work):
        assert task.status == task.S_OK
        assert task.num_restarts == i
        assert task.num_corrections == 0

    assert relax_work[1].input["dilatmx"] == 1.03

    # check that when decreasing the dilatmx it actually takes the previously relaxed
    # structure and does not start from scratch again: the lattice should not be the same.
    assert not np.allclose(relax_work.ion_task.get_final_structure().lattice_vectors(),
                           relax_work.ioncell_task.input.structure.lattice_vectors())

    flow.rmtree()



#!/usr/bin/env python
from __future__ import print_function, unicode_literals, division, absolute_import
import sys

from abipy.data import AbinitFilesGenerator


class MyGenerator(AbinitFilesGenerator):
    """This class generates the output files used in the unit tests and in the examples."""
    # Subclasses must define the following class attributes:
    # List of pseudos (basenames in abipy/data/pseudos)
    pseudos = ["14si.pspnc"]

    # Mapping old_name --> new_name for the output files that must be preserved.
    files_to_save = {
        "out_DS4_SIGRES.nc": "si_g0w0ppm_nband10_SIGRES.nc",
        "out_DS5_SIGRES.nc": "si_g0w0ppm_nband20_SIGRES.nc",
        "out_DS6_SIGRES.nc": "si_g0w0ppm_nband30_SIGRES.nc",
    }


if __name__ == "__main__":
    sys.exit(MyGenerator().run())



"""This module contains lookup table with the name of the ABINIT variables."""
from __future__ import annotations

import os
import warnings
import numpy as np
from typing import Union

from pprint import pformat #, pprint
from monty.string import is_string, boxed
from monty.functools import lazy_property
from monty.termcolor import cprint
from pymatgen.core.units import bohr_to_ang
from abipy.core.structure import Structure, dataframes_from_structures
from abipy.core.mixins import Has_Structure, TextFile, NotebookWriter
from abipy.abio.abivar_database.variables import get_codevars

__all__ = [
    "is_abivar",
    "is_abiunit",
    "AbinitInputFile",
    "AbinitInputParser",
    "structure_from_abistruct_fmt",
]


def is_anaddb_var(varname: str) -> bool:
    """True if varname is a valid Anaddb variable."""
    return varname in get_codevars()["anaddb"]


def is_abivar(varname: str) -> bool:
    """True if s is an ABINIT variable."""
    # Add include statement
    # FIXME: These variables should be added to the database.
    extra = ["include", "xyzfile"]
    return varname in get_codevars()["abinit"] or varname in extra


# TODO: Move to new directory
ABI_OPERATORS = set(["sqrt", ])

ABI_UNIT_NAMES = {
    s.lower() for s in (
        "au", "nm",
        "Angstr", "Angstrom", "Angstroms", "Bohr", "Bohrs",
        "eV", "Ha", "Hartree", "Hartrees", "K", "Ry", "Rydberg", "Rydbergs",
        "T", "Tesla",)
}


def is_abiunit(s: str) -> bool:
    """
    True if string is one of the units supported by the ABINIT parser
    """
    if not is_string(s): return False
    return s.lower() in ABI_UNIT_NAMES


def expand_star_syntax(s: str) -> str:
    """
    Evaluate star syntax. Return new string
    Remember that Abinit does not accept white spaces.
    For example `typat 2 * 1` is not valid.

    >>> assert expand_star_syntax("3*2") == '2 2 2'
    >>> assert expand_star_syntax("2 *1") == '1 1'
    >>> assert expand_star_syntax("1 2*2") == '1 2 2'
    >>> assert expand_star_syntax("*2") == '*2'
    """
    s = s.strip()
    if "*" not in s:
        return s
    else:
        # Handle e.g `pawecutdg*`
        if s[0].isalpha() and s[-1] == "*": return s

    s = s.replace("*", " * ").strip()
    tokens = s.split()
    #tokens = [c.rstrip().lstrip() for c in s.split()]

    # Handle "*2" case i.e. return "*2"
    if len(tokens) == 2 and tokens[0] == "*":
        assert tokens[1] != "*"
        return "".join(tokens)

    #print(s, tokens)
    l = []
    while tokens:
        c = tokens.pop(0)
        if c == "*":
            num = int(l.pop(-1))
            val = tokens.pop(0)
            l.extend(num * [val])
        else:
            l.append(c)

    return " ".join(l)


def str2array_bohr(obj):
    if not is_string(obj):
        return np.asarray(obj)

    # Treat e.g. acell 3 * 1.0
    obj = expand_star_syntax(obj)
    # Numpy does not understand "0.00d0 0.00d0"
    obj = obj.lower().replace("d", "e")

    tokens = obj.split()
    if not tokens[-1].isalpha():
        # No unit
        return np.fromstring(obj, sep=" ")

    unit = tokens[-1]
    if unit in ("angstr", "angstrom", "angstroms"):
        return np.fromstring(" ".join(tokens[:-1]), sep=" ") / bohr_to_ang
    elif unit in ("bohr", "bohrs", "au"):
        return np.fromstring(" ".join(tokens[:-1]), sep=" ")
    else:
        raise ValueError("Don't know how to handle unit: %s" % str(unit))


def str2array(obj, dtype=float) -> np.ndarray:
    if not is_string(obj): return np.asarray(obj)
    if obj.startswith("*"):
        raise ValueError("This case should be treated by the caller: %s" % str(obj))
    s = expand_star_syntax(obj)
    # Numpy does not understand "0.00d0 0.00d0"
    s = s.lower().replace("d", "e")
    return np.fromstring(s, sep=" ", dtype=dtype)


def eval_abinit_operators(tokens: list[str]) -> list[str]:
    """
    Receive a list of strings, find the occurences of operators supported
    in the input file (e.g. sqrt), evalute the expression and return new list of strings.

    .. note:

        This function is not recursive hence expr like sqrt(1/2) are not supported
    """
    import math # noqa: F401
    import re
    re_sqrt = re.compile(r"[+|-]?sqrt\((.+)\)")

    values = []
    for tok in tokens:
        if tok.startswith("'") or tok.startswith('"'):
            values.append(tok)
        else:
            m = re_sqrt.match(tok)
            if m:
                tok = tok.replace("sqrt", "math.sqrt")
                tok = str(eval(tok))
            if "/" in tok:
                tok = str(eval(tok))
            values.append(tok)

    return values


def abi_tokenize(string: str) -> list[str]:

    string = string.strip()

    # stack = [(char, [start, end]), ...] where char is " or '
    stack = []
    for i, c in enumerate(string):
        if c in ("'", '"'):
            if stack and stack[-1][1][1] == -1:
                # Close quotation mark. Note that the algo is not very robust as nested quotation marks are not supported!
                entry = stack[-1]
                if c != entry[0]:
                    raise ValueError("Found unclosed quotation marks in: %s" % string)
                entry = stack[-1][1][1] = i
            else:
                # First hit.
                stack.append((c, [i, -1]))

    if not stack:
        # No ' or " found in string.
        return string.split()

    # Unroll stack and buid tokens so that e.g. "foo bar" is treated as a single token.
    tokens = []
    base = 0
    for item in stack:
        start, stop = item[1][0], item[1][1]
        # Consistency check:
        if stop == -1:
            raise ValueError("Found unclosed quotation mark in: %s" % string)
        if start > 0:
            tokens.extend(string[base:start-1].strip().split())
            base = stop + 1
        tokens.append(string[start:stop + 1])

    if base < len(string):
        tokens.append(string[base:])

    return tokens


class Dataset(dict, Has_Structure):

    @lazy_property
    def structure(self) -> Structure:
        """
        The initial structure associated to the dataset.
        """

        # First of all check whether the structure is defined through an external file.
        if "structure" in self:
            s = self["structure"].replace('"', "")
            filetype, path = s.split(":")
            #print("filetype:", filetype)
            from abipy import abilab
            if filetype == "poscar":
                return abilab.Structure.from_file(path)

            elif filetype == "abivars":
                with open(path, "rt") as fh:
                    return structure_from_abistruct_fmt(fh.read())

            else:
                try:
                    with abilab.abiopen(path) as abifile:
                        if hasattr(abifile, "final_structure"):
                            return abifile.final_structure
                        else:
                            return abifile.structure
                except Exception as exc:
                    raise RuntimeError("Error while opening file: `%s`:\n%s" % (path, exc))

        # Get lattice.
        kwargs = {}
        if "angdeg" in self:
            if "rprim" in self:
                raise ValueError("rprim and angdeg cannot be used together!")
            angdeg = str2array(self["angdeg"])
            angdeg.shape = (3)
            kwargs["angdeg"] = angdeg
        else:
            # Handle structure specified with rprim.
            kwargs["rprim"] = str2array_bohr(self.get("rprim", "1.0 0 0 0 1 0 0 0 1"))

        # Default value for acell.
        acell = str2array_bohr(self.get("acell", "1.0 1.0 1.0"))

        # Get important dimensions.
        ntypat = int(self.get("ntypat", 1))
        natom = int(self.get("natom", 1))

        # znucl(npsp)
        znucl = self["znucl"]
        if znucl.startswith("*"):
            i = znucl.find("*")
            znucl_size = natom if "npsp" not in self else int(self["npsp"])
            znucl = znucl_size * [float(znucl[i+1:])]
        else:
            znucl = str2array(self["znucl"])

        # v67mbpt/Input/t12.in
        typat = self["typat"]
        if typat.startswith("*"):
            i = typat.find("*")
            typat = np.array(natom * [int(typat[i+1:])], dtype=int)
        else:
            typat = str2array(self["typat"], dtype=int)

        # Extract atomic positions.
        # Select first natom entries (needed if multidatasets with different natom)
        #    # v3/Input/t05.in
        typat = typat[:natom]
        for k in ("xred", "xcart", "xangst"):
            toarray = str2array_bohr if k == "xcart" else str2array
            if k in self:
                arr = np.reshape(toarray(self[k]), (-1, 3))
                kwargs[k] = arr[:natom]
                break
        else:
            raise ValueError("xred|xcart|xangst must be given in input")

        try:
            return Structure.from_abivars(acell=acell, znucl=znucl, typat=typat, **kwargs)
        except Exception as exc:
            print("Wrong inputs passed to Structure.from_abivars:")
            print("acell:", acell, "znucl:", znucl, "typat:", typat, "kwargs:", kwargs, sep="\n")
            raise exc

    def get_vars(self) -> dict:
        """
        Return dictionary with variables. The variables describing the crystalline structure
        are removed from the output dictionary.
        """
        geovars = {"acell", "angdeg", "rprim", "ntypat", "natom", "znucl", "typat", "xred", "xcart", "xangst"}
        return {k: self[k] for k in self if k not in geovars}

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, post=None, mode="text", verbose=0) -> str:
        """
        String representation.

        Args:
            post: String that will be appended to the name of the variables
            mode: Either `text` or `html` if HTML output with links is wanted.
            verbose: Verbosity level.
        """
        post = post if post is not None else ""
        if mode == "html":
            from abipy.abio.abivars_db import get_abinit_variables
            var_database = get_abinit_variables()

        lines = []
        app = lines.append
        for k in sorted(list(self.keys())):
            vname = k + post
            if mode == "html": vname = var_database[k].html_link(label=vname)
            app("%s %s" % (vname, str(self[k])))

        return "\n".join(lines) if mode == "text" else "\n".join(lines).replace("\n", "<br>")

    def _repr_html_(self):
        """Integration with jupyter_ notebooks."""
        return self.to_string(mode="html")


class AbinitInputFile(TextFile, Has_Structure, NotebookWriter):
    """
    This object parses the Abinit input file, stores the variables in
    dict-like objects (Datasets) and build `Structure` objects from
    the input variables. Mainly used for inspecting the structure
    declared in the Abinit input file.
    """

    @classmethod
    def from_string(cls, string: str) -> AbinitInputFile:
        """Build the object from string."""
        import tempfile
        _, filename = tempfile.mkstemp(suffix=".abi", text=True)
        with open(filename, "wt") as fh:
            fh.write(string)
        return cls(filename)

    def __init__(self, filepath: str):
        super().__init__(filepath)

        with open(filepath, "rt") as fh:
            self.string = fh.read()

        self.datasets = AbinitInputParser().parse(self.string)
        self.ndtset = len(self.datasets)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []
        app = lines.append
        header = 10 * "=" + " Input File " + 10 * "="
        app(header)
        app(self.string)
        app(len(header) * "=" + "\n")

        # Print info on structure(s).
        if self.structure is not None:
            app(self.structure.spget_summary())
        else:
            structures = [dt.structure for dt in self.datasets]
            app("Input file contains %d structures:" % len(structures))
            for i, structure in enumerate(structures):
                app(boxed("Dataset: %d" % (i+1)))
                app(structure.spget_summary())
                app("")

            dfs = dataframes_from_structures(structures, index=[i+1 for i in range(self.ndtset)])
            app(boxed("Tabular view (each row corresponds to a dataset structure)"))
            app("")
            app("Lattice parameters:")
            app(str(dfs.lattice))
            app("")
            app("Atomic positions:")
            app(str(dfs.coords))

        return "\n".join(lines)

    @lazy_property
    def has_multi_structures(self) -> bool:
        """True if input defines multiple structures."""
        return self.structure is None

    def _repr_html_(self):
        """Integration with jupyter notebooks."""
        from abipy.abio.abivars_db import repr_html_from_abinit_string
        return repr_html_from_abinit_string(self.string)
        #return self.to_string(mode="html"))

    def close(self) -> None:
        """NOP, required by ABC."""

    @lazy_property
    def structure(self) -> Structure:
        """
        The structure defined in the input file.

        If the input file contains multiple datasets **AND** the datasets
        have different structures, this property returns None.
        In this case, one has to access the structure of the individual datasets.
        For example:

            input.datasets[0].structure

        gives the structure of the first dataset.
        """
        for dt in self.datasets[1:]:
            if dt.structure != self.datasets[0].structure:
                warnings.warn("Datasets have different structures. Returning None. Use input.datasets[i].structure")
                return None

        return self.datasets[0].structure

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        if not self.has_multi_structures:
            yield self.structure.plot(show=False)
            yield self.structure.plot_bz(show=False)
        else:
            for dt in self.datasets:
                yield dt.structure.plot(show=False)
                yield dt.structure.plot_bz(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write an ipython notebook to nbpath. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("abinp = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(abinp)"),
        ])

        if self.has_multi_structures:
            nb.cells.extend([
                nbv.new_code_cell("""\
for dataset in inp.datasets:
    print(dataset.structure)"""),
            ])

        if self.ndtset > 1:
            nb.cells.extend([
                nbv.new_code_cell("""\
for dataset in abinp.datasets:
    print(dataset)"""),
            ])

        return self._write_nb_nbpath(nb, nbpath)

    def get_differences(self, other, ignore_vars=None) -> list[str]:
        """
        Get the differences between this AbinitInputFile and another.
        """
        diffs = []
        to_ignore = {"acell", "angdeg", "rprim", "ntypat", "natom", "znucl", "typat", "xred", "xcart", "xangst"}
        if ignore_vars is not None:
            to_ignore.update(ignore_vars)
        if self.ndtset != other.ndtset:
            diffs.append(f"Number of datasets in this file is {self.ndtset} "
                         f"while other file has {other.ndtset} datasets.")
            return diffs
        for idataset, self_dataset in enumerate(self.datasets):
            other_dataset = other.datasets[idataset]
            if self_dataset.structure != other_dataset.structure:
                diffs.append("Structures are different.")
            self_dataset_dict = dict(self_dataset)
            other_dataset_dict = dict(other_dataset)
            for k in to_ignore:
                if k in self_dataset_dict:
                    del self_dataset_dict[k]
                if k in other_dataset_dict:
                    del other_dataset_dict[k]
            common_keys = set(self_dataset_dict.keys()).intersection(other_dataset_dict.keys())
            self_only_keys = set(self_dataset_dict.keys()).difference(other_dataset_dict.keys())
            other_only_keys = set(other_dataset_dict.keys()).difference(self_dataset_dict.keys())
            if self_only_keys:
                diffs.append(f"The following variables are in this file but not in other: "
                             f"{', '.join([str(k) for k in self_only_keys])}")
            if other_only_keys:
                diffs.append(f"The following variables are in other file but not in this one: "
                             f"{', '.join([str(k) for k in other_only_keys])}")
            for k in common_keys:
                if self_dataset_dict[k] != other_dataset_dict[k]:
                    diffs.append(f"The variable '{k}' is different in the two files:\n"
                                 f" - this file:  '{self_dataset_dict[k]}'\n"
                                 f" - other file: '{other_dataset_dict[k]}'")
        return diffs


class AbinitInputParser:

    verbose = 0

    def parse(self, s: str):
        """
        This function receives a string `s` with the Abinit input and return
        a list of :class:`Dataset` objects.
        """
        # Remove "=" and comments from lines.
        lines = []
        s = s.replace("=", " ")
        for line in s.splitlines():
            line.strip()
            i = line.find("#")
            if i != -1: line = line[:i]
            i = line.find("!")
            if i != -1: line = line[:i]
            if line: lines.append(line)

        # 1) Build string of the form `var1 value1 var2 value2`
        # 2) Split string in tokens.
        # 3) Evaluate star syntax i.e. "3*2" ==> '2 2 2'
        # 4) Evaluate operators e.g. sqrt(0.75)
        #
        # NB: Step 3 is needed because we are gonna use python to evaluate the operators and
        #     in abinit `2*sqrt(0.75)` means `sqrt(0.75) sqrt(0.75)` and not math multiplication!
        #
        #tokens = " ".join(lines).split()
        #print("tokens:\n", pformat(tokens))

        tokens = abi_tokenize(" ".join(lines))  #print("new tokens:\n", pformat(tokens))
        if self.verbose: print("tokens", tokens)

        # Step 3-4
        new_tokens = []
        for t in tokens:
            if t.startswith("'") or t.startswith('"'):
                new_tokens.append(t)
            else:
                l = expand_star_syntax(t).split() # ;print(l)
                new_tokens.extend(l)

        tokens = new_tokens
        if self.verbose: print("new_tokens", new_tokens)

        tokens = self.eval_abinit_operators(tokens) # ; print(tokens)

        varpos = []
        for pos, tok in enumerate(tokens):
            #if not isnewvar(ok): continue
            #print("token:", tok)

            if tok[0].isalpha():
                #print("tok[0]", tok[0])
                # Either new variable or string defining the unit or operator e.g. sqrt
                if is_abiunit(tok) or tok in ABI_OPERATORS or "?" in tok:
                    continue

                # Got new variable
                if tok[-1].isdigit(): # and "?" not in tok:
                    # Handle dataset index.
                    l = []
                    for i, c in enumerate(tok[::-1]):
                        if c.isalpha(): break
                        l.append(c)
                    else:
                        raise ValueError("Cannot find dataset index in token: %s" % tok)
                    l.reverse()
                    #if not is_abivar(tok):
                        #continue
                        #raise ValueError("Expecting variable but got: %s" % tok)

                #print("new varname `", tok, "` at positions", pos)
                varpos.append(pos)

        varpos.append(len(tokens))

        # Build dict {varname --> value_string}
        dvars = {}
        for i, pos in enumerate(varpos[:-1]):
            varname = tokens[pos]  # ; print("varname:", varname)
            if pos + 2 == len(tokens):
                dvars[varname] = tokens[-1]
            else:
                dvars[varname] = " ".join(tokens[pos+1: varpos[i+1]])
        # print("dvars:\n", pformat(dvars))

        err_lines = []
        for k, v in dvars.items():
            if not v:
                err_lines.append("key `%s` was not parsed correctly (empty value)" % k)
        if err_lines:
            raise RuntimeError("\n".join(err_lines))

        # Get value of ndtset.
        ndtset = int(dvars.pop("ndtset", 1))
        udtset = dvars.pop("udtset", None)
        jdtset = dvars.pop("jdtset", None)
        if udtset is not None:
            raise NotImplementedError("udtset is not supported")

        # Build list of datasets.
        datasets = [Dataset() for i in range(ndtset)]

        # Treat all variables without a dataset index
        kv_list = list(dvars.items())
        for k, v in kv_list:
            if k[-1].isdigit() or any(c in k for c in ("?", ":", "+", "*")): continue
            for d in datasets: d[k] = v
            dvars.pop(k)

        # Treat all variables with a dataset index except those with "?", ":", "+"
        kv_list = list(dvars.items())
        for k, v in kv_list:
            if any(c in k for c in ("?", ":", "+", "*")): continue
            varname, idt = self.varname_dtindex(k)
            dvars.pop(k)
            #if varname == "angdeg": raise ValueError("got angdeg")
            if idt > ndtset:
                if self.verbose: print("Ignoring key: %s because ndtset: %d" % (k, ndtset))
                continue
            datasets[idt-1][varname] = v

        # Now treat series e.g. ecut: 10 ecut+ 5 (NB: ? is not treated here)
        kv_list = list(dvars.items())
        for k, v in kv_list:
            if "?" in k: continue
            if ":" not in k: continue
            # TODO units
            vname = k[:-1]
            start = str2array(dvars.pop(k))

            # Handle ecut+ or ecut*
            incr = dvars.pop(vname + "+", None)
            if incr is not None:
                incr = str2array(incr)
                for dt in datasets:
                    dt[vname] = start.copy()
                    start += incr

            else:
                mult = dvars.pop(vname + "*")
                mult = str2array(mult)
                for dt in datasets:
                    dt[vname] = start.copy()
                    start *= mult

        # Consistency check
        # 1) dvars should be empty
        if dvars:
            raise ValueError("Don't know how handle variables in:\n%s" % pformat(dvars), indent=4)

        # 2) Keys in datasets should be valid Abinit input variables.
        wrong = []
        for i, dt in enumerate(datasets):
            wlist = [k for k in dt if not is_abivar(k)]
            if wlist:
                wrong.extend(("dataset %d" % i, wlist))
        if wrong:
            raise ValueError("Found variables that are not registered in the abipy database:\n%s" % pformat(wrong, indent=4))

        # 3) We don't support spg builder: dataset.structure will fail or, even worse,
        #    spglib will segfault so it's better to raise here!
        for dt in datasets:
            if "spgroup" in dt or "nobj" in dt:
                raise NotImplementedError(
                    "Abinit spgroup builder is not supported. Structure must be given explicitly!")

        if jdtset is not None:
            # Return the datasets selected by jdtset.
            datasets = [datasets[i-1] for i in np.fromstring(jdtset, sep=" ", dtype=int)]

        return datasets

    @staticmethod
    def eval_abinit_operators(tokens):
        return eval_abinit_operators(tokens)

    @staticmethod
    def varname_dtindex(tok):
        """
        >>> p = AbinitInputParser()
        >>> assert p.varname_dtindex("acell1") == ("acell", 1)
        >>> assert p.varname_dtindex("fa1k2") == ("fa1k", 2)
        """
        l = []
        for i, c in enumerate(tok[::-1]):
            if c.isalpha(): break
            l.append(c)
        else:
            raise ValueError("Cannot find dataset index in: %s" % tok)

        assert i > 0
        l.reverse()
        dtidx = int("".join(l))
        varname = tok[:len(tok)-i]

        return varname, dtidx


def format_string_abivars(varname, value, code="abinit") -> str:
    """
    Given a variable and its value, check if it is of string type and wraps it
    in double quotes if needed.

    Args:
        varname: the name of the variable.
        value: the value to format.
        code: the code the variable refers to (e.g. "abinit", "anaddb")

    Returns:
        The properly formatted value.
    """

    var = get_codevars()[code].get(varname)
    if var and var.vartype == "string":
        if not isinstance(value, (list, tuple)):
            value = [value]

        str_values = [str(v).strip() for v in value]

        # handle gruns_ddbs separately as it supports a different syntax
        # incompatible with other variables
        if varname == "gruns_ddbs":
            join_str = "\n    "
            for i, str_val in enumerate(str_values):
                if str_val[0] != '"':
                    str_values[i] = '"' + str_val
                if str_val[-1] != '"':
                    str_values[i] += '"'
        else:
            join_str = ",\n    "
            if str_values[0][0] != '"':
                str_values[0] = '"' + str_values[0]
            if str_values[-1][-1] != '"':
                str_values[-1] += '"'

        if len(str_values) > 1:
            str_values[0] = "\n    " + str_values[0]

        return join_str.join(str_values)
    return value


def structure_from_abistruct_fmt(string):
    """
    Parse geometrical information given in the structure:abivars format return Structure object

    A typical input file in "structure:abivars" format looks like::

        # MgB2 lattice structure.
        natom   3
        acell   2*3.086  3.523 Angstrom
        rprim   0.866025403784439  0.5  0.0
               -0.866025403784439  0.5  0.0
                0.0                0.0  1.0

        # Atomic positions
        xred_symbols
            0.0  0.0  0.0 Mg
            1/3  2/3  0.5 B
            2/3  1/3  0.5 B

    i.e. the zcucl, typat and ntypat vars are implicitly defined via the xred_symbols table.
    """
    # Algorithm is as follows:
    #     1) parse the xred_symbols section using a record-based parser.
    #     2) Compute xred, znucl, ntypat and typat from the xred_symbols section
    #     3) Add these values in string format to the head to get a "standard" Abinit input file.
    #     4) Init structure from the "standard" Abinit input file.
    found = 0
    head = []
    xred, symbols = [], []

    for line in string.split("\n"):
        line = line.strip()

        if line.lower() == "xred_symbols":
            found += 1
            continue
        if not found:
            head.append(line)
        if found and line:
            tokens = line.split() #; print(line)
            xred.append([float(t) for t in eval_abinit_operators(tokens[:3])])
            symbols.append(tokens[3])

    if not xred:
        raise ValueError("Cannot find xred_symbols section in string:\n %s" % string)

    xred = np.reshape(xred, (-1, 3))

    # Set small values to zero. This usually happens when the CIF file
    # does not give structure parameters with enough digits.
    xred = np.where(np.abs(xred) > 1e-8, xred, 0.0)

    # Compute ntypat, typat, znucl from xred and symbols.
    natom = len(xred)
    ntypat = 1
    typat = [0]
    for iatom in range(1, natom):
        found = -1
        for jj in range(ntypat):
            if symbols[iatom] == symbols[typat[jj]]:
                found = jj
                break
        if found == -1:
            ntypat += 1
            typat.append(ntypat - 1)
        else:
            typat.append(found)

    from pymatgen.core.periodic_table import Element
    znucl = np.empty(ntypat)
    for iatom in range(natom):
        itypat = typat[iatom]
        znucl[itypat] = Element(symbols[iatom]).Z

    # Convert C indexing to Fortran.
    typat = np.array(typat) + 1

    # Build standard Abinit input string.
    lines = []
    app = lines.append
    app("ntypat %d" % ntypat)
    app("typat " + " ".join("%d" % t for t in typat))
    app("znucl " + " ".join("%f" % z for z in znucl))
    app("xred ")
    for iatom in range(natom):
        v = xred[iatom]
        lines.append("%.16f %.16f %.16f" % (v[0], v[1], v[2]))

    s = "\n".join(head) + "\n" + "\n".join(lines) #; print(s)

    return AbinitInputFile.from_string(s).structure


def validate_input_parser(abitests_dir=None, input_files=None):
    """
    validate/test AbinitInput parser.

    Args:
        dirpath: Abinit tests directory.
        input_files: List of Abinit input files.

    Return: Exit code.
    """
    def is_abinit_input(path):
        """
        True if path is one of the input files used in the Abinit Test suite.
        """
        if path.endswith(".abi"): return True
        if not path.endswith(".in"): return False

        with open(path, "rt") as fh:
            for line in fh:
                if "executable" in line and "abinit" in line: return True
            return False

    # Files are collected in paths.
    paths = []

    if abitests_dir is not None:
        print("Analyzing directory %s for input files" % abitests_dir)

        for dirpath, dirnames, filenames in os.walk(abitests_dir):
            for fname in filenames:
                path = os.path.join(dirpath, fname)
                if is_abinit_input(path): paths.append(path)

            #import ast
            #init_path = os.path.join(dirpath, "__init__.py")
            #with open(init_path, "rt") as f:
            #    source = f.read()
            #    start = source.find("inp_files = [")
            #    if start == -1:
            #        print("ignoring ", init_path)
            #        continue
            #    stop = source.find("]", start)
            #    if stop == -1:
            #        raise ValueError("Invalid code in %s" % init_path)
            #    print(init_path)
            #    inp_basenames = ast.literal_eval(source[start:stop+1])
            #    print(int_basenames)

    if input_files is not None:
        print("Analyzing files ", str(input_files))
        for arg in input_files:
            if is_abinit_input(arg): paths.append(arg)

    nfiles = len(paths)
    if nfiles == 0:
        cprint("Empty list of input files.", "red")
        return 0

    print("Found %d Abinit input files" % len(paths))
    errpaths = []
    for path in paths:
        print(path + ": ", end="")
        try:
            inp = AbinitInputFile.from_file(path)
            s = str(inp)
            cprint("OK", "green")
        except Exception as exc:
            if not isinstance(exc, NotImplementedError):
                cprint("FAILED", "red")
                errpaths.append(path)
                import traceback
                print(traceback.format_exc())
                #print("[%s]: Exception:\n%s" % (path, str(exc)))
                #with open(path, "rt") as fh:
                #    print(10*"=" + "Input File" + 10*"=")
                #    print(fh.read())
                #    print()
            else:
                cprint("NOTIMPLEMENTED", "magenta")

    if errpaths:
        cprint("failed: %d/%d [%.1f%%]" % (len(errpaths), nfiles, 100 * len(errpaths)/nfiles), "red")
        for i, epath in enumerate(errpaths):
            cprint("[%d] %s" % (i, epath), "red")
    else:
        cprint("All input files successfully parsed!", "green")

    return len(errpaths)



"""
"""
from __future__ import annotations

import os

from pprint import pformat
from typing import Dict, Any, Literal, List
from ruamel import yaml
from pydantic import Field, root_validator, PrivateAttr
from abipy.core.structure import Structure
from abipy.abio.inputs import AbinitInput
from abipy.abio.abivars import is_abivar
from abipy.abio.factories import gs_input, ebands_input
from abipy.htc.base_models import AbipyModel
from abipy.htc.pseudos_models import PseudoSpecs


class MetaParam(AbipyModel):

    name: str
    exclude_abivars: List[str] = []
    require_abivars: List[str] = []
    exclude_meta: List[str] = []
    require_meta: List[str] = []


_ALL_METAPARAMS = {
    "kppa": MetaParam(name="kppa", exclude_abivars=["nkpt", "ngkpt", "kptrlatt"]),
    "qppa": MetaParam(name="qppa"),
}


class _Specs(AbipyModel):

    meta_params: Dict[str, Any] = Field(
            default_factory=dict,
            description="Dictionary with meta-parameters")

    abivars: Dict[str, Any] = Field(
            default_factory=dict,
            description="Dictionary with extra Abinit variables that will be added to the AbinitInput")

    spin_mode: str = Field(..., description="Spin polarization mode")

    # Private attributes
    #_smearing: str = PrivateAttr()

    def __init__(self, **data):

        abivars = data["abivars"]
        nsppol = abivars.pop("nsppol", 1)
        nspinor = abivars.pop("nspinor", 1)
        nspden = abivars.pop("nspden", 1)

        from pymatgen.io.abinit.abiobjects import _mode2spinvars  #  SpinMode
        for spin_mode, var in _mode2spinvars.items():
            if (nsppol == var.nsppol and nspinor == var.nspinor and nspden == var.nspden):
                break
        else:
            raise ValueError(f"Cannot find spin_mode associated to\n"
                             f"nsppol: {nsppol}, nspinor: {nspinor}, nspden: {nspden}")

        #from pymatgen.io.abinit.abiobjects import Smearing
        #occopt = int(abivars.pop("occopt", 1))
        #tsmear = abivars.pop("tsmear", 0.01)
        #tsmear = float(tsmear)
        #smearing = Smearing(occopt, tsmear)

        super().__init__(spin_mode=spin_mode, **data)
        self.validate_instance()


    def __str__(self) -> str:
        lines = []
        app = lines.append
        app(f"Spec: {self.__class__.__name__}")
        app(f"Doc: {self.__class__.__doc__}")
        app("")
        app(f"meta_params:\n {pformat(self.meta_params)}\n")
        app(f"abivars:\n {pformat(self.abivars)}\n")

        return "\n".join(lines)

    def validate_instance(self):
        cls_name = self.__class__.__name__

        for meta_name in self.meta_params:
            meta = _ALL_METAPARAMS.get(meta_name, None)
            if meta is None:
                raise ValueError(f"In {cls_name}: unknown meta_parameter `{meta_name}`")

            if meta.exclude_abivars:
                for aname in self.abivars:
                    if aname in meta.exclude_abivars:
                        raise ValueError(f"In {cls_name}: metavariable `{meta_name}` "
                                         f"is not compatible with Abinit variable: `{aname}`")

            for aname in meta.require_abivars:
                if aname not in self.abivars:
                    raise ValueError(f"In {cls_name}: metavariable `{meta_name}` "
                                     f"requires Abinit variable: `{aname}`")

            for mp in self.meta_params:
                if mp in meta.exclude_meta:
                    raise ValueError(f"In {cls_name}: metavariable `{meta_name}` "
                                     f"is not compatible with other meta parameter: `{mp}`")

            for mp in meta.require_meta:
                raise ValueError(f"In {cls_name}: metavariable `{meta_name}` requires `{mp}`")

        if "fband" and "nband" in self.abivars:
            raise ValueError("fband and nband cannot be used together.")


    @root_validator
    def check_abivars(cls, values):
        """
        Make sure the names specified in the abivars section are valid.
        """
        for k in values.get("abivars"):
            if not is_abivar(k):
                raise ValueError(f"Unknown Abinit variable: `{k}`")
        return values

    @root_validator
    def check_meta_params(cls, values):
        """
        Check that the meta_params are:

        - registered
        - do not ext

        """
        abivars = values.get("abivars")
        meta_params = values.get("meta_params")

        # Check whether meta_param is supported by this class
        for meta_name in meta_params:
            if meta_name not in cls._supported_meta_params:
                raise ValueError(f"Specs `{cls.__name__}` does not support meta_parameter: `{meta_name}`")

        for meta_name in meta_params:
            meta = _ALL_METAPARAMS.get(meta_name, None)
            if meta is None:
                raise ValueError(f"Unknown meta_parameter: `{meta_name}`")

            #if meta.exclude_abivars and any(vname in meta.exclude_abivars
            #meta.require_abivars
            #meta.exclude_meta
            #meta.require_meta
            #raise ValueError(f"Unknown Abinit variable: `{k}`")

        return values


class GsScfSpecs(_Specs):
    """
    Specifications for ground-state SCF calculations
    """

    _supported_meta_params = [
        "kppa",
    ]



class GsNscfKpathSpecs(_Specs):
    """
    Specifications for ground-state NSCF calculations with k-path (used to compute band structures)
    """

    _supported_meta_params = [
    ]


class GsNscfKdosSpecs(_Specs):
    """
    Specifications for ground-state NSCF calculations with k-mesh (used to compute e-DOS)
    """

    _supported_meta_params = [
        "kppa",
    ]


class RelaxSpecs(_Specs):
    """
    Specifications for Structural relaxations
    """

    _supported_meta_params = [
        "kppa",
    ]


class PhononSpecs(_Specs):
    """
    Specifications for phonon calculations within DFPT
    """

    _supported_meta_params = [
        "qppa",
    ]


KNOWN_SPECS = [
        "gs_scf_specs",
        "gs_nscf_kpath_specs",
        "gs_nscf_kdos_specs",
        "relax_specs",
        #"phonon_specs",
]


# Absolute path to the directory with yaml files.
PROTO_DIRPATH = os.path.join(os.path.dirname(__file__), "protocols")


class Protocol(AbipyModel):
    """
    A protocol contains meta parameters used to generate input files
    """
    info: str = Field(..., description="Human readable string with the description of the protocol.")

    #name: str =  Field(..., description=".")

    #md5: str = Field(..., description="")

    accuracy: Literal["low", "normal", "high"] = Field(..., description="")

    pseudos_specs: PseudoSpecs = Field(..., description="Pseudopotential Specs")

    gs_scf_specs: GsScfSpecs = Field(..., description="Specs for GS SCF calculation")

    gs_nscf_kpath_specs: GsNscfKpathSpecs = Field(
            None,
            description="Specs for NSCF band structure calculations along a k-path")

    gs_nscf_kdos_specs: GsNscfKdosSpecs = Field(
            None,
            description="Specs for NSCF calculations with k-mesh to compute DOS")

    relax_specs: RelaxSpecs = Field(
            None,
            description="Specs for structural relaxations")

    # validators
    #_normalize_name = validator('name', allow_reuse=True)(normalize)
    #_normalize_name = reuse_validator('name')(normalize)

    #@root_validator
    #def check_pseudo_specs(cls, values):
    #    pseudos_specs = values.get("pseudos_specs")
    #    return values

    @classmethod
    def from_file(cls, filepath: str) -> Protocol:
        """Build a protocol object from a Yaml file."""
        with open(filepath, "rt") as fh:
            data = yaml.safe_load(fh)
            return cls.parse_data(data)

    @classmethod
    def from_yaml_string(cls, yaml_string: str) -> Protocol:
        """Build a protocol object from a string in Yaml format."""
        data = yaml.safe_load(yaml_string)
        return cls.parse_data(data)

    @classmethod
    def from_name(cls, name: str) -> Protocol:
        """Build a protocol object from the name of the Yaml file in the AbiPy library."""
        filepath = os.path.join(PROTO_DIRPATH, name)
        return cls.from_file(filepath)

    @classmethod
    def get_all_abipy_protocols(cls) -> List[Protocol]:
        """
        Return list of all registered protocols.
        """
        yaml_files = [f for f in os.listdir(PROTO_DIRPATH) if f.endswith(".yml")]

        protocols = []
        for y in yaml_files:
            proto = cls.from_file(os.path.join(PROTO_DIRPATH, y))
            protocols.append(proto)

        return protocols

    @classmethod
    def parse_data(cls, data: dict) -> Protocol:
        """
        Parse dictionary and return Protocol.
        """
        data = data.copy()

        #name = data.pop("name", None)
        #if name is None:
        #    raise ValueError(f"Cannot find `name` section in data")

        info = data.pop("info", None)
        if info is None:
            raise ValueError(f"Cannot find `info` key in document")

        accuracy = data.pop("accuracy", None)
        if accuracy is None:
            raise ValueError(f"Cannot find `accuracy` section in document")

        # Get global abinit variables first.
        global_abivars = data.pop("global_abivars", {})

        # Get pseudo specifications and build model.
        d = data.pop("pseudos_specs", None)
        if d is None:
            raise ValueError(f"Cannot find `pseudos_specs` section in data")
        pseudos_specs = PseudoSpecs.from_repo_table_name(d["repo_name"], d["table_name"])

        specs = {}
        for spec_name in KNOWN_SPECS:
            specs[spec_name] = data.pop(spec_name, None)

        if data:
            raise ValueError(f"Found unknown sections with keys: `{list(data.keys())}`")

        # Add global variables
        # This means that one can always override the value per entry
        extend_map = {}
        for spec_name, d in specs.items():
            if d is None: continue
            for n in ["meta_params", "abivars"]:
                if n not in d: d[n] = {}
            new_dict = global_abivars.copy()
            new_dict.update(d["abivars"])
            d["abivars"] = new_dict
            extends = d.pop("extends", None)
            if extends:
                extend_map[spec_name] = str(extends)

        # Implement "extends" syntax.
        for spec_name, super_name in extend_map.items():
            super_d = d[super_name]
            #new_metaps = super_d["meta_params"].copy()
            #new_metaps.update(d[spec_name]["meta_params"])
            #d[spec_name]["meta_params"] = new_metaps
            #new_abivars = super_d["abivars"].copy()
            #new_abivars.update(d[spec_name]["abivars"])
            #d[spec_name]["abivars"] = new_abivars

        return cls(info=info, accuracy=accuracy, pseudos_specs=pseudos_specs, **specs)

    def __init__(self, **data):
        super().__init__(**data)
        self.validate_instance()

    def validate_instance(self):
        cls_name = self.__class__.__name__
        #raise ValueError(f"In {cls_name}: unknown meta_parameter `{meta_name}`")

    #def __repr__(self) -> str:
    #    return "<>"

    def __str__(self) -> str:
        lines = [f"info:\n{self.info}"]
        app = lines.append
        app(f"pseudos: {self.pseudos_specs}")

        for spec_name in KNOWN_SPECS:
            spec = getattr(self, spec_name)
            app(f"spec_name: {spec_name}")
            app(f"spec: {spec}")
            app("")

        return "\n".join(lines)

    ####################
    # Factory functions.
    ####################

    def get_gs_scf_input(self, structure: Structure) -> AbinitInput:
        """
        Build and return an input for a GS SCF calculation for the given structure
        """
        pseudos = self.pseudos_specs.get_pseudos()
        scf_specs = self.gs_scf_specs

        scf_inp = gs_input(structure, pseudos,
                           #kppa=scf_specs.kppa,
                           ecut=None, pawecutdg=None, scf_nband=None,
                           accuracy=self.accuracy,
                           spin_mode=scf_specs.spin_mode,
                           smearing="fermi_dirac:0.1 eV",
                           charge=0.0,
                           )

        # Add extra abivars
        scf_inp.set_vars(**scf_specs.abivars)

        return scf_inp

    def get_ebands_input(self, structure: Structure):
        """
        Build and return list of inputs for a GS SCF + NSCF for the given structure
        """
        pseudos = self.pseudos_specs.get_pseudos()

        scf_specs = self.gs_scf_specs
        nscf_specs = self.gs_nscf_kpath_specs
        if nscf_specs is None:
            raise ValueError("band structure calculation requires the specification of `gs_nscf_kpath_specs`")

        multi = ebands_input(structure, pseudos,
                             #kppa=self.kppa,
                             nscf_nband=None,
                             ndivsm=self.ndivsm,
                             ecut=None, pawecutdg=None,
                             scf_nband=None,
                             accuracy=self.accuracy,
                             spin_mode=self.spin_mode,
                             smearing=self.smearing,
                             charge=self.charge,
                             #dos_kppa=self.dos_kppa,
                             )

        #nscf_inp = scf_inp.make_ebands_input(ndivsm=15, tolwfr=1e-20, nscf_nband=None, nb_extra=10)
        scf_inp, nscf_inp = multi.split_datasets()

        # Add extra abivars
        scf_inp.set_vars(**self.gs_scf_specs.abivars)
        nscf_inp.set_vars(**self.gs_nscf_specs.abivars)

        return scf_inp, nscf_inp

    #def get_relax_input(self, structure: Structure) -> AbinitInput:
    #    """
    #    Build and return an input to relax the input structure
    #    """
    #    scf_inp = self.get_gs_scf_input(structure)
    #    relax_specs = self.relax_specs
    #    if relax_specs is None:
    #         raise ValueError("structure relaxations require the specifications of `relax_specs`")
    #    relax_input = None
    #    return relax_input


if __name__ == "__main__":
    proto = Protocol.from_file("protocols/NC-PBE-SR-PDv0.4.yml")
    print(proto)

    from abipy.data.ucells import structure_from_ucell
    structures = [structure_from_ucell(name) for name in ("Si",)] # "Si-shifted")]
    for structure in structures:
        scf_inp = proto.get_gs_scf_input(structure)
        print(scf_inp)
